<?php

namespace App\Utils\Paypal;

use App\Exceptions\PaypalExceptions;
use App\Models\Cart;
use App\Models\Paypal;
use Constantine\Pay\PayFacade;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Support\Str;

class PaypalApiService implements PaypalService
{
    protected Model $model;

    public function __construct(Paypal $paypal)
    {
        $this->model = $paypal;
    }

    /**
     * @note 生成paypal amount items配置 存放产品信息
     *
     * @param string $userId 购物车id
     **/
    public function createItemsAmount(string $userId): array
    {
        $items = [];
        $amount = [];
        $total = 0;
        $tax = 0;
        $subtotal = 0;
        $currency = config('paypal.currency');
        $carts = Cart::query()->where('session_id', $userId)->where('is_buy', 0)->get();
        if (sizeof($carts) < 1) {
           abort(500,'Por favor adicione o produto');
        }
        foreach ($carts as $cart) {
            $taxUnit = 0;
            $item = [];
            $product = $cart->product;
            $item["name"] = $cart->brand . ' ' . $cart->jian . ' ' . env('APP_URL');
            $item['sku'] = $product->pcode;
            $item['unit_amount'] = [
                'currency_code' => $currency,
                'value' => $product->price,
            ];
            $item['tax'] = [
                'currency_code' => $currency,
                'value' => 0.00,
            ];
            $item['quantity'] = $cart->numbers;
            $item['description'] = $product->pcode . '/' . env('APP_URL') . "buy($cart->numbers)" . ';';
            $item['category'] = 'PHYSICAL_GOODS';
            $items[] = $item;
            $subtotal = bcadd($subtotal, bcmul($product->price, $cart->numbers, 4), 2);

        }

        $tax = bcmul($subtotal, env('paypal_tax', 0), 2);
        $shipping =0;
        $total = bcadd(bcadd($subtotal,$tax,4),$shipping,2);

        $amount = [
            'currency_code' => $currency,
            'value' => "$total",
            'breakdown' =>
                [
                    'item_total' => [
                        'currency_code' => $currency,
                        'value' => "$subtotal",
                    ],

                    'shipping' => [
                        'currency_code' => $currency,
                        'value' => $shipping
                    ],
                    'handling' => [
                        'currency_code' => $currency,
                        'value' => '0.00'
                    ],
                    'tax_total' => [
                        'currency_code' => $currency,
                        'value' => "$tax"
                    ],
                    'shipping_discount' => [
                        'currency_code' => $currency,
                        'value' => '0.00'
                    ],
                ],
        ];
        return compact('items', 'amount');
    }

    /**
     *
     * @note 生成用户购买地址
     *
     * @param array $info 用户信息
     **/
    public function createShipping(array $info = []): array
    {
        if (!$info) {
            return [];
        }
        return [
            'method' => 'United States Postal Service',
            'name' =>
                [
                    'full_name' => $info['name'],
                ],
            'address' =>
                [
                    'address_line_1' => $info['address1'],
                    'address_line_2' => $info['address2'] ?? '',
                    'admin_area_2' => $info['city'],
                    'admin_area_1' => $info['state'] ?? '',
                    'postal_code' => $info['postal'],
                    'country_code' => $info['country'],
                ]
        ];
    }

    /**
     * @note 生成付款配置文件
     **/

    public function createConfigData(string $userId, $returnUrl, $cancelUrl, array $info = []): array
    {
        $config = $this->createItemsAmount($userId);
        $items = $config['items'];
        $amount = $config['amount'];
        $shipping = $this->createShipping($info);
        if ($shipping) {

            return [
                'intent' => 'CAPTURE',
                'application_context' => [
                    'return_url' => $returnUrl,
                    'cancel_url' => $cancelUrl,
                    'brand_name' => config('paypal.brand_name'),
                    'locale' => config('paypal.locale'),
                    'landing_page' => 'BILLING',
                    'user_action' => 'PAY_NOW',
                    'shipping_preference' => 'SET_PROVIDED_ADDRESS',
                ],
                'purchase_units' => [
                    0 => [
                        'reference_id' => 'PUHF',
                        'description' => 'Sporting Goods',
                        'custom_id' => 'CUST-HighFashions',
                        'soft_descriptor' => 'HighFashions',
                        'amount' => $amount,
                        'items' => $items,
                        'shipping' => $shipping
                    ]
                ],

            ];
        }
        $data = [
            'intent' => 'CAPTURE',
            'application_context' => [
                'return_url' => $returnUrl,
                'cancel_url' => $cancelUrl,
                'brand_name' => config('paypal.brand_name'),
                'locale' => config('paypal.locale'),
                'landing_page' => 'BILLING',
                'user_action' => 'PAY_NOW',
                'shipping_preference' => 'GET_FROM_FILE',
            ],
            'purchase_units' => [
                0 => [
                    'reference_id' => 'PUHF',
                    'description' => 'Sporting Goods',
                    'custom_id' => 'CUST-HighFashions',
                    'soft_descriptor' => 'HighFashions',
                    'amount' => $amount,
                    'items' => $items,
                ]
            ],
        ];
        return $data;
    }

    /**
     * @note 付款成功后数据入库
     * @note 具体参数可查看paypal文档 https://developer.paypal.com/docs/api/orders/v2/#orders_capture
     *
     * @param array $response paypal server 返回的数据
     * @param string $userId 用户id
     *
     **/
    public function handleResponse(array $response, string $userId)
    {
        /**
         * @var $data ['trans_id'] paypal交易号
         * @var $data ['payer_id'] paypal用户id
         * @var $data ['local_id'] 本地生成订单id
         * @var $data ['order_id'] 订单id可查询订单
         *
         **/
        $data = [];
        $carts = Cart::query()->with('product')->where('session_id', $userId)->where('is_buy', 0)
            ->get();
        $description = '';
        $shipping = $response['purchase_units'][0]['shipping'];
        $amount = $response['purchase_units'][0]['payments']['captures'][0]['amount'];
        $data['session_id'] = $userId;
        $data['trans_id'] = $response['purchase_units'][0]['payments']['captures'][0]['id'];
        $data['payer_id'] = $response['payer']['payer_id'];
        $data['local_id'] = now()->format('Ymd') . Str::random(8);
        $data['order_id'] = $response['id'];
        $data['name'] = $shipping['name']['full_name'];
        $data['phone'] = $response['payment_source']['paypal']['phone_number']['national_number'];
        $data['email'] = $response['payment_source']['paypal']['email_address'];
        $data['postal'] = $shipping['address']['postal_code'];
        $data['country'] = $shipping['address']['country_code'];
        $data['state'] = $shipping['address']['admin_area_1']??'';
        $data['city'] = $shipping['address']['admin_area_2'];
        $data['street1'] = $shipping['address']['address_line_1'];
        $data['street2'] = $shipping['address']['address_line_2'] ?? '';
        $data['ip'] = request()->ip();
        $data['total'] = $amount['value'];
        $data['currency'] = $amount['currency_code'];

        if (app()->environment() === 'production') {
            Cart::query()->whereIn('id', $carts->pluck('id')->toArray())->update([
                'is_buy' => 1
            ]);
        }

        $provider = new \Srmklive\PayPal\Services\PayPal();
        $paypalToken = $provider->getAccessToken();
        $descriptionResponse = $provider->showOrderDetails($data['order_id']);
        $items = $descriptionResponse['purchase_units'][0]['items'];
        foreach ($items as $item) {
            $description .= $item['description'] . '|';
        }
        $data['des'] = $description;
        $this->model->create($data);
//        PayFacade::orderShippedEvent($carts, $data);
    }

    /**
     *
     * 根据数据库产品创建paypal产品
     **/
    public function createProduct($item): array
    {
        return [
            "name" => trim($item->title),
            "description" => trim($item->title),
            "type" => "SERVICE",
            "category" => "SOFTWARE",
            "image_url" => trim($item->okpcode),
            "home_url" => trim($item->url),
        ];
    }

    public function updateProduct(array $product): array
    {
        $data = [];

        unset($product['type']);
        unset($product['category']);
        foreach ($product as $k => $v) {
            $data[] = $this->generateUpdateData($k, $v);
        }
        return $data;
    }

    public function generateUpdateData(string $column, string $value, string $op = 'replace')
    {
        return [
            'op' => $op,
            'path' => '/' . $column,
            'value' => $value
        ];
    }
}
