<?php

namespace App\Utils\Paypal;

use App\Models\Cart;
use App\Models\Paypal;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Str;
use PayPal\Api\Amount;
use PayPal\Api\Details;
use PayPal\Api\Item;
use PayPal\Api\ItemList;
use PayPal\Api\Payer;
use PayPal\Api\Payment;
use PayPal\Api\PaymentExecution;
use PayPal\Api\Presentation;
use PayPal\Api\RedirectUrls;
use PayPal\Api\Transaction;
use PayPal\Api\WebProfile;
use PayPal\Auth\OAuthTokenCredential;
use PayPal\Rest\ApiContext;

class PaypalSdkService implements PaypalService
{
    protected object $apiContext;
    protected $currency;
    protected Model $model;

    public function __construct(Model $model)
    {

        $this->model = $model;

        $paypalModel = env('PAYPAL_MODE');

        if ($paypalModel === 'sandbox') {
            $clientId = config('paypal.sandbox.client_id');
            $clientSecret = config('paypal.sandbox.client_secret');
        } else {
            $clientId = config('paypal.live.client_id');
            $clientSecret = config('paypal.live.client_secret');
        }
        $this->currency = env('PAYPAL_CURRENCY');
        $config = [
            'mode' => $paypalModel,
            'log.LogEnabled' => true,
            'log.FileName' => storage_path('logs/paypal-sdk.log'),
            'log.LogLevel' => $paypalModel === 'sandbox' ? 'DEBUG' : 'INFO', // PLEASE USE `INFO` LEVEL FOR LOGGING IN LIVE ENVIRONMENTS
            'cache.enabled' => true,
            //'cache.FileName' => '/PaypalCache' // for determining paypal cache directory
            'http.CURLOPT_CONNECTTIMEOUT' => 60
            // 'http.headers.PayPal-Partner-Attribution-Id' => '123123123'
            //'log.AdapterFactory' => '\PayPal\Log\DefaultLogFactory' // Factory class implementing \PayPal\Log\PayPalLogFactory
        ];
        $this->apiContext = new ApiContext(new OAuthTokenCredential($clientId, $clientSecret));
        $this->apiContext->setConfig($config);
    }

    /**
     *
     * 创建支付
     *
     **/
    public function createConfigData(string $userId, $returnUrl, $cancelUrl, array $info = [])
    {

        /**
         * @var  $shipping 运费
         * @var $subtotal 小计
         * @var  $tax 税费
         * @var  $total 总价
         *
         **/
        $shipping = 0;
        $subtotal = 0;
        $tax = 0;
        $total = 0;
        $items = [];
        $carts = Cart::query()->with('product')->where('session_id', $userId)->where('is_buy', 0)->get();
        $payer = new Payer();
        $payer->setPaymentMethod("paypal");

        foreach ($carts as $cart) {
            $name = $cart->product->cs . " " . $cart->jian . ' ' . getenv('APP_URL');
            $time = now();
            $des = $cart->product->pcode . '-' . 'buy(' . $cart->numbers . ')' . $cart->product->cs . ' ' . $cart->jian; // 商品描述
            $quantity = $cart->numbers; //数量
            $subtotal += $cart->product->price * $cart->numbers; //除去运费的总价
            $item = new Item();
            $item->setName($name) //名字
            ->setDescription($des) //描述
            ->setCurrency($this->currency) //货币单位
            ->setQuantity($quantity) //数量
            ->setSku($cart->product->pcode) // 产品信息
            ->setPrice($cart->product->price); //价格
            $items[] = $item;
        }
        //设置物品item
        $itemList = new ItemList();
        $itemList->setItems($items);
        //设置运费 税费
        $details = new Details();
        $details->setShipping($shipping)
            ->setTax($tax)->setSubtotal($subtotal);
        //设置总价
        $total = $subtotal + $shipping + $tax;
        $amount = new Amount();
        $amount->setCurrency($this->currency)->setTotal($total)->setDetails($details);

        $transaction = new Transaction();
        $transaction->setAmount($amount)
            ->setItemList($itemList)
            ->setDescription("Payment description")
            ->setInvoiceNumber(uniqid());
        //设置支付成功取消支付地址
//        $baseUrl = getBaseUrl();
        $redirectUrls = new RedirectUrls();
        $redirectUrls->setReturnUrl($returnUrl)
            ->setCancelUrl($cancelUrl);

        $payment = new Payment();
        //paypal 商户体验设置
        $webProfile = new WebProfile();
        $presentation = new Presentation();
        $presentation->setBrandName(config('paypal.brand_name'))->setLocaleCode(config('paypal.locale'))
            ->setLogoImage(env('APP_URL') . '/' . 'favicon.ico');
        $webProfile->setName(config('paypal.brand_name') . uniqid())->setPresentation($presentation);
        $webProfileId = $webProfile->create($this->apiContext)->getId();

        $payment->setIntent("sale")
            ->setPayer($payer)
            ->setRedirectUrls($redirectUrls)
            ->setTransactions(array($transaction))
            ->setExperienceProfileId($webProfileId);
        $request = clone $payment;
        try {
            $payment->create($this->apiContext);
        } catch (Exception $ex) {
            Log::error($ex->getMessage() . $ex->getCode());
        }
//        $approvalUrl = $payment->getApprovalLink();
        return $payment;
    }

    public function handleResponse(array $params, string $userId)
    {
        $paymentId = $params['paymentId'];
        $payerId = $params['PayerID'];
        $payment = Payment::get($paymentId, $this->apiContext);
        $execution = new PaymentExecution();
        $execution->setPayerId($payerId);
        $result = $payment->execute($execution, $this->apiContext);
        $data['email'] = $result->payer->payer_info->email;
        $data['phone'] = $result->payer->payer_info->phone;
        $data['street1'] = $result->payer->payer_info->shipping_address->line1;
        $data['street2'] = $result->payer->payer_info->shipping_address->line2;
        $data['name'] = $result->payer->payer_info->shipping_address->recipient_name;
        $data['country'] = $result->payer->payer_info->shipping_address->country_code;
        $data['trans_id'] = $result->transactions[0]->related_resources[0]->sale->id;
        $data['payer_id'] = $result->payer->payer_info->payer_id;
        $data['city'] = $result->payer->payer_info->shipping_address->city;
        $data['state'] = $result->payer->payer_info->shipping_address->state;
        $data['postal'] = $result->payer->payer_info->shipping_address->postal_code;
        $data['total'] = $result->transactions[0]->amount->total;
        $items = $result->transactions[0]->item_list->items;
        $description = '';
        foreach ($items as $k => $v) {
            $description .= $v->description;
        }

        $data['local_id'] = now()->format('Ymd') . Str::random(6);
        $data['ip'] = request()->ip();
        $data['des'] = $description;
        $data['session_id'] = session()->getId();
        $data['created_at'] = now();
        $data['currency'] = env('PAYPAL_CURRENCY');
        $this->model->create($data);
        $carts = Cart::query()->where('session_id', $userId)
            ->where('is_buy', 0)->update([
                'is_buy' => 1
            ]);

    }
}
